import numpy as np
import pytest

import autofit as af
from autolens.lens.util import lens_fit_util as util
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing, lens_fit
from autolens.model.hyper import hyper_data as hi
from autolens.model.galaxy import galaxy as g
from autolens.model.inversion import inversions
from autolens.model.inversion import pixelizations
from autolens.model.inversion import regularization
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.unit.mock.model.mock_profiles import MockLightProfile


class TestLensProfileFit:
    class TestLikelihood:
        def test__1x2_image__no_psf_blurring__tracing_fits_data_with_chi_sq_5(self):
            # The image plane image generated by the galaxy is [1.0, 1.0]

            # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

            psf = abstract_data.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )

            ccd_data = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf,
                noise_map=np.ones((3, 4)),
            )
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            lens_data_7x7 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensProfileFit(lens_data=lens_data_7x7, tracer=tracer)

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (
                fit.mask(return_in_2d=True)
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.noise_map_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.noise_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.model_image_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.model_image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.residual_map_1d == np.array([4.0, 3.0])).all()
            assert (
                fit.residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.normalized_residual_map_1d == np.array([4.0, 3.0])).all()
            assert (
                fit.normalized_residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.chi_squared_map_1d == np.array([16.0, 9.0])).all()
            assert (
                fit.chi_squared_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 9.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__1x2_image__include_psf_blurring__tracing_fits_data_with_chi_sq_4(
            self
        ):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            psf = abstract_data.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )

            ccd_data = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf,
                noise_map=np.ones((3, 4)),
            )
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            lens_data_7x7 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensProfileFit(lens_data=lens_data_7x7, tracer=tracer)

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (
                fit.mask(return_in_2d=True)
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.noise_map_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.noise_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.model_image_1d == np.array([1.0, 4.0])).all()
            assert (
                fit.model_image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.residual_map_1d == np.array([4.0, 0.0])).all()
            assert (
                fit.residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.normalized_residual_map_1d == np.array([4.0, 0.0])).all()
            assert (
                fit.normalized_residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.chi_squared_map_1d == np.array([16.0, 0.0])).all()
            assert (
                fit.chi_squared_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 16.0
            assert fit.reduced_chi_squared == 16.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                16.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test_hyper_galaxy_changes_noise_above_from_1_to_2__reflected_in_likelihood(
            self
        ):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            # The hyper_galaxy galaxy increases the noise in both pixels by 1.0, to 2.0.

            # This reduces the chi squared to 2.0 instead of 4.0

            psf = abstract_data.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )

            ccd_data = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf,
                noise_map=np.ones((3, 4)),
            )
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            lens_data_7x7 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(
                redshift=0.5,
                light_profile=MockLightProfile(value=1.0, size=2),
                hyper_galaxy=g.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image_1d=np.ones(2),
                hyper_galaxy_image_1d=np.ones(2),
                hyper_minimum_value=0.0,
            )

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensProfileFit(lens_data=lens_data_7x7, tracer=tracer)

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (
                fit.mask(return_in_2d=True)
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.noise_map_1d == np.array([2.0, 2.0])).all()
            assert (
                fit.noise_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.model_image_1d == np.array([1.0, 4.0])).all()
            assert (
                fit.model_image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.residual_map_1d == np.array([4.0, 0.0])).all()
            assert (
                fit.residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.normalized_residual_map_1d == np.array([2.0, 0.0])).all()
            assert (
                fit.normalized_residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.chi_squared_map_1d == np.array([4.0, 0.0])).all()
            assert (
                fit.chi_squared_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 4.0
            assert fit.reduced_chi_squared == 4.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.likelihood == -0.5 * (4.0 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0))

        def test__hyper_image_changes_background_sky__reflected_in_likelihood(self):

            psf = abstract_data.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )

            ccd_data = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf,
                noise_map=np.ones((3, 4)),
            )
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            lens_data_7x7 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            hyper_image_sky = hi.HyperImageSky(sky_scale=1.0)

            fit = lens_fit.LensProfileFit(
                lens_data=lens_data_7x7, tracer=tracer, hyper_image_sky=hyper_image_sky
            )

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (
                fit.mask(return_in_2d=True)
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (fit.image_1d == np.array([6.0, 5.0])).all()
            assert (
                fit.image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 6.0, 5.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.noise_map_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.noise_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.model_image_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.model_image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.residual_map_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.normalized_residual_map_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.normalized_residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.chi_squared_map_1d == np.array([25.0, 16.0])).all()
            assert (
                fit.chi_squared_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 25.0, 16.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 41.0
            assert fit.reduced_chi_squared == 41.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                41.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__hyper_background_changes_background_noise_map__reflected_in_likelihood(
            self
        ):

            psf = abstract_data.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )

            ccd_data = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf,
                noise_map=np.ones((3, 4)),
            )
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            lens_data_7x7 = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=1)

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            hyper_background_noise = hi.HyperBackgroundNoise(noise_scale=1.0)

            fit = lens_fit.LensProfileFit(
                lens_data=lens_data_7x7,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (
                fit.mask(return_in_2d=True)
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.noise_map_1d == np.array([2.0, 2.0])).all()
            assert (
                fit.noise_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.model_image_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.model_image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.residual_map_1d == np.array([4.0, 3.0])).all()
            assert (
                fit.residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.normalized_residual_map_1d == np.array([2.0, 1.5])).all()
            assert (
                fit.normalized_residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 1.5, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.chi_squared_map_1d == np.array([4.0, 2.25])).all()
            assert (
                fit.chi_squared_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 2.25, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 6.25
            assert fit.reduced_chi_squared == 6.25 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                6.25 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0)
            )

        def test__hyper_noise_map_max_changes_noise_map__reflected_in_likelihood(self):

            psf = abstract_data.PSF(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scale=1.0,
            )

            ccd_data = ccd.CCDData(
                5.0 * np.ones((3, 4)),
                pixel_scale=1.0,
                psf=psf,
                noise_map=np.ones((3, 4)),
            )
            ccd_data.image[1, 2] = 4.0

            mask = msk.Mask(
                array=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scale=1.0,
            )

            lens_data_7x7 = ld.LensData(
                ccd_data=ccd_data, mask=mask, sub_grid_size=1, hyper_noise_map_max=1.0
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = g.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            hyper_background_noise = hi.HyperBackgroundNoise(noise_scale=1.0)

            fit = lens_fit.LensProfileFit(
                lens_data=lens_data_7x7,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert (fit.mask_1d == np.array([False, False])).all()
            assert (
                fit.mask(return_in_2d=True)
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (fit.image_1d == np.array([5.0, 4.0])).all()
            assert (
                fit.image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.noise_map_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.noise_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.model_image_1d == np.array([1.0, 1.0])).all()
            assert (
                fit.model_image(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.residual_map_1d == np.array([4.0, 3.0])).all()
            assert (
                fit.residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.normalized_residual_map_1d == np.array([4.0, 3.0])).all()
            assert (
                fit.normalized_residual_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (fit.chi_squared_map_1d == np.array([16.0, 9.0])).all()
            assert (
                fit.chi_squared_map(return_in_2d=True)
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 9.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.likelihood == -0.5 * (
                25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

    class TestCompareToManual:
        def test___all_lens_fit_quantities__no_hyper_methods(self, lens_data_7x7):

            g0 = g.Galaxy(
                redshift=0.5,
                light_profile=lp.EllipticalSersic(intensity=1.0),
                mass_profile=mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = g.Galaxy(
                redshift=1.0, light_profile=lp.EllipticalSersic(intensity=1.0)
            )

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0, g1], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            assert lens_data_7x7.noise_map(return_in_2d=True) == pytest.approx(
                fit.noise_map(return_in_2d=True)
            )

            model_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_7x7.convolver_image
            )
            model_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=model_image_1d
            )

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_7x7.image_1d,
                mask=lens_data_7x7.mask_1d,
                model_data=model_image_1d,
            )
            residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=residual_map_1d
            )

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

            normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=lens_data_7x7.noise_map_1d,
            )

            normalized_residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=normalized_residual_map_1d
            )

            assert normalized_residual_map_1d == pytest.approx(
                fit.normalized_residual_map_1d, 1e-4
            )
            assert normalized_residual_map_2d == pytest.approx(
                fit.normalized_residual_map(return_in_2d=True)
            )

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=lens_data_7x7.noise_map_1d,
            )

            chi_squared_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=chi_squared_map_1d
            )

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True)
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_7x7.mask_2d
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                noise_map=lens_data_7x7.noise_map(return_in_2d=True),
                mask=lens_data_7x7.mask_2d,
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
            assert likelihood == fit.figure_of_merit

        def test___lens_fit_galaxy_image_dict__corresponds_to_blurred_galaxy_images(
            self, lens_data_7x7
        ):
            g0 = g.Galaxy(
                redshift=0.5,
                light_profile=lp.EllipticalSersic(intensity=1.0),
                mass_profile=mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = g.Galaxy(
                redshift=1.0, light_profile=lp.EllipticalSersic(intensity=1.0)
            )
            g2 = g.Galaxy(redshift=1.0)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0, g1, g2], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            g0_image_plane_image_1d = g0.intensities_from_grid(
                grid=lens_data_7x7.grid_stack.sub,
                return_in_2d=False,
                return_binned=True,
            )
            g0_image_plane_bluring_image_1d = g0.intensities_from_grid(
                grid=lens_data_7x7.grid_stack.blurring,
                return_in_2d=False,
                return_binned=False,
            )

            g0_blurred_image_plane_image_1d = lens_data_7x7.convolver_image.convolve_image(
                image_array=g0_image_plane_image_1d,
                blurring_array=g0_image_plane_bluring_image_1d,
            )

            g1_image_plane_image_1d = g1.intensities_from_grid(
                grid=tracer.source_plane.grid_stack.sub,
                return_in_2d=False,
                return_binned=True,
            )
            g1_image_plane_bluring_image_1d = g1.intensities_from_grid(
                grid=tracer.source_plane.grid_stack.blurring,
                return_in_2d=False,
                return_binned=False,
            )

            g1_blurred_image_plane_image_1d = lens_data_7x7.convolver_image.convolve_image(
                image_array=g1_image_plane_image_1d,
                blurring_array=g1_image_plane_bluring_image_1d,
            )

            assert fit.galaxy_image_1d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_1d, 1.0e-4
            )
            assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_1d, 1.0e-4
            )
            assert (fit.galaxy_image_1d_dict[g2] == np.zeros(9)).all()

            assert fit.model_image_1d == pytest.approx(
                fit.galaxy_image_1d_dict[g0] + fit.galaxy_image_1d_dict[g1], 1.0e-4
            )

            g0_blurred_image_plane_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=g0_blurred_image_plane_image_1d
            )
            g1_blurred_image_plane_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=g1_blurred_image_plane_image_1d
            )

            assert fit.galaxy_image_2d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_2d, 1.0e-4
            )
            assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_2d, 1.0e-4
            )

            assert fit.model_image(return_in_2d=True) == pytest.approx(
                fit.galaxy_image_2d_dict[g0] + fit.galaxy_image_2d_dict[g1], 1.0e-4
            )

        def test___all_lens_fit_quantities__including_hyper_methods(
            self, lens_data_7x7
        ):

            hyper_noise_map_max = 0.2
            lens_data_7x7.hyper_noise_map_max = hyper_noise_map_max

            hyper_image_sky = hi.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = hi.HyperBackgroundNoise(noise_scale=1.0)

            image_1d = hyper_image_sky.image_scaled_sky_from_image(
                image=lens_data_7x7.image_1d
            )

            hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
                noise_map=lens_data_7x7.noise_map_1d
            )

            g0 = g.Galaxy(
                redshift=0.5,
                light_profile=lp.EllipticalSersic(intensity=1.0),
                mass_profile=mp.SphericalIsothermal(einstein_radius=1.0),
                hyper_galaxy=g.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image_1d=np.ones(9),
                hyper_galaxy_image_1d=np.ones(9),
                hyper_minimum_value=0.0,
            )
            g1 = g.Galaxy(
                redshift=1.0, light_profile=lp.EllipticalSersic(intensity=1.0)
            )

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0, g1], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
                noise_map_1d=lens_data_7x7.noise_map_1d
            )

            hyper_noise_map_1d = hyper_noise_map_background_1d + hyper_noise_1d

            hyper_noise_map_1d[
                hyper_noise_map_1d > hyper_noise_map_max
            ] = hyper_noise_map_max

            hyper_noise_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=hyper_noise_map_1d
            )

            assert hyper_noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert hyper_noise_map_2d == pytest.approx(fit.noise_map(return_in_2d=True))

            model_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_7x7.convolver_image
            )

            model_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=model_image_1d
            )

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=image_1d, mask=lens_data_7x7.mask_1d, model_data=model_image_1d
            )
            residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=residual_map_1d
            )

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

            normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=hyper_noise_map_1d,
            )

            normalized_residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=normalized_residual_map_1d
            )

            assert normalized_residual_map_1d == pytest.approx(
                fit.normalized_residual_map_1d, 1e-4
            )
            assert normalized_residual_map_2d == pytest.approx(
                fit.normalized_residual_map(return_in_2d=True)
            )

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=hyper_noise_map_1d,
            )

            chi_squared_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=chi_squared_map_1d
            )

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True)
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_7x7.mask_2d
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                noise_map=hyper_noise_map_2d, mask=lens_data_7x7.mask_2d
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)
            assert likelihood == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
            self, lens_data_7x7
        ):

            g0 = g.Galaxy(
                redshift=0.5,
                light_profile=lp.EllipticalSersic(intensity=1.0),
                mass_profile=mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = g.Galaxy(
                redshift=1.0, light_profile=lp.EllipticalSersic(intensity=1.0)
            )

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0, g1], image_plane_grid_stack=lens_data_7x7.grid_stack
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            blurred_profile_image_2d_of_planes = tracer.blurred_profile_image_plane_image_2d_of_planes_from_convolver_image(
                convolver_image=lens_data_7x7.convolver_image
            )

            assert (
                blurred_profile_image_2d_of_planes[0] == fit.model_image_2d_of_planes[0]
            ).all()
            assert (
                blurred_profile_image_2d_of_planes[1] == fit.model_image_2d_of_planes[1]
            ).all()

            unmasked_blurred_profile_image = tracer.unmasked_blurred_profile_image_plane_image_from_psf(
                psf=lens_data_7x7.psf
            )

            assert (
                unmasked_blurred_profile_image
                == fit.unmasked_blurred_profile_image_plane_image
            ).all()

            unmasked_blurred_profile_image_of_planes = tracer.unmasked_blurred_profile_image_plane_image_of_planes_from_psf(
                psf=lens_data_7x7.psf
            )

            assert (
                unmasked_blurred_profile_image_of_planes[0]
                == fit.unmasked_blurred_profile_image_plane_image_of_planes[0]
            ).all()
            assert (
                unmasked_blurred_profile_image_of_planes[1]
                == fit.unmasked_blurred_profile_image_plane_image_of_planes[1]
            ).all()

            unmasked_blurred_profile_image_of_galaxies = tracer.unmasked_blurred_profile_image_plane_image_of_plane_and_galaxies_from_psf(
                psf=lens_data_7x7.psf
            )

            assert (
                unmasked_blurred_profile_image_of_galaxies[0][0]
                == fit.unmasked_blurred_profile_image_plane_image_of_planes_and_galaxies[
                    0
                ][
                    0
                ]
            ).all()
            assert (
                unmasked_blurred_profile_image_of_galaxies[1][0]
                == fit.unmasked_blurred_profile_image_plane_image_of_planes_and_galaxies[
                    1
                ][
                    0
                ]
            ).all()


class TestLensInversionFit:
    class TestCompareToManual:
        def test___all_lens_fit_quantities__no_hyper_methods(self, lens_data_7x7):

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)

            g0 = g.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g.Galaxy(redshift=0.5), g0],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )
            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image_1d=lens_data_7x7.image_1d,
                noise_map_1d=lens_data_7x7.noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
            )

            assert inversion.reconstructed_data_1d == pytest.approx(
                fit.model_image_1d, 1e-4
            )
            assert inversion.reconstructed_data_2d == fit.model_image(return_in_2d=True)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_7x7.image_1d,
                mask=lens_data_7x7.mask_1d,
                model_data=inversion.reconstructed_data_1d,
            )
            residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=residual_map_1d
            )

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

            normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=lens_data_7x7.noise_map_1d,
            )

            normalized_residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=normalized_residual_map_1d
            )

            assert normalized_residual_map_1d == pytest.approx(
                fit.normalized_residual_map_1d, 1e-4
            )
            assert normalized_residual_map_2d == pytest.approx(
                fit.normalized_residual_map(return_in_2d=True)
            )

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=lens_data_7x7.noise_map_1d,
            )
            chi_squared_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=chi_squared_map_1d
            )

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True)
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_7x7.mask_2d
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_7x7.mask_2d,
                noise_map=lens_data_7x7.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4
            )

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___lens_fit_galaxy_image_dict__has_inversion_reconstructed_data(
            self, lens_data_7x7
        ):
            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)

            g0 = g.Galaxy(redshift=0.5)
            g1 = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0, g1],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image_1d=lens_data_7x7.image_1d,
                noise_map_1d=lens_data_7x7.noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
            )

            assert (fit.galaxy_image_1d_dict[g0] == np.zeros(9)).all()

            assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
                inversion.reconstructed_data_1d, 1.0e-4
            )

            assert fit.model_image_1d == pytest.approx(
                fit.galaxy_image_1d_dict[g1], 1.0e-4
            )

            assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
                inversion.reconstructed_data_2d, 1.0e-4
            )

            assert fit.model_image(return_in_2d=True) == pytest.approx(
                fit.galaxy_image_2d_dict[g1], 1.0e-4
            )

        def test___all_lens_fit_quantities__include_hyper_methods(self, lens_data_7x7):

            hyper_noise_map_max = 0.2
            lens_data_7x7.hyper_noise_map_max = hyper_noise_map_max

            hyper_image_sky = hi.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = hi.HyperBackgroundNoise(noise_scale=1.0)

            image_1d = hyper_image_sky.image_scaled_sky_from_image(
                image=lens_data_7x7.image_1d
            )

            hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
                noise_map=lens_data_7x7.noise_map_1d
            )

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)

            g0 = g.Galaxy(
                redshift=0.5,
                pixelization=pix,
                regularization=reg,
                hyper_galaxy=g.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image_1d=np.ones(9),
                hyper_galaxy_image_1d=np.ones(9),
                hyper_minimum_value=0.0,
            )

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g.Galaxy(redshift=0.5), g0],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
                noise_map_1d=lens_data_7x7.noise_map_1d
            )
            hyper_noise_map_1d = hyper_noise_map_background_1d + hyper_noise_1d
            hyper_noise_map_1d[
                hyper_noise_map_1d > hyper_noise_map_max
            ] = hyper_noise_map_max

            hyper_noise_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=hyper_noise_map_1d
            )
            assert hyper_noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert hyper_noise_map_2d == pytest.approx(fit.noise_map(return_in_2d=True))

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )
            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image_1d=image_1d,
                noise_map_1d=hyper_noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
            )

            assert inversion.reconstructed_data_1d == pytest.approx(
                fit.model_image_1d, 1e-4
            )
            assert inversion.reconstructed_data_2d == fit.model_image(return_in_2d=True)

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=image_1d,
                mask=lens_data_7x7.mask_1d,
                model_data=inversion.reconstructed_data_1d,
            )
            residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=residual_map_1d
            )

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

            normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=hyper_noise_map_1d,
            )

            normalized_residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=normalized_residual_map_1d
            )

            assert normalized_residual_map_1d == pytest.approx(
                fit.normalized_residual_map_1d, 1e-4
            )
            assert normalized_residual_map_2d == pytest.approx(
                fit.normalized_residual_map(return_in_2d=True)
            )

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=hyper_noise_map_1d,
            )
            chi_squared_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=chi_squared_map_1d
            )

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True)
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_7x7.mask_2d
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_7x7.mask_2d, noise_map=hyper_noise_map_2d
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4
            )

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
            self, lens_data_7x7
        ):

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)

            g0 = g.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g.Galaxy(redshift=0.5), g0],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image_1d=lens_data_7x7.image_1d,
                noise_map_1d=lens_data_7x7.noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
            )

            assert (fit.model_image_2d_of_planes[0] == np.zeros((7, 7))).all()
            assert inversion.reconstructed_data_2d == pytest.approx(
                fit.model_image_2d_of_planes[1], 1.0e-4
            )


class TestLensProfileInversionFit:
    class TestCompareToManual:
        def test___all_lens_fit_quantities__no_hyper_methods(self, lens_data_7x7):
            galaxy_light = g.Galaxy(
                redshift=0.5, light_profile=lp.EllipticalSersic(intensity=1.0)
            )

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[galaxy_light, galaxy_pix],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_7x7.convolver_image
            )

            blurred_profile_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=blurred_profile_image_1d
            )

            assert blurred_profile_image_1d == pytest.approx(
                fit.blurred_profile_image_1d, 1e-4
            )
            assert blurred_profile_image_2d == pytest.approx(
                fit.blurred_profile_image(return_in_2d=True)
            )

            profile_subtracted_image_1d = (
                lens_data_7x7.image_1d - blurred_profile_image_1d
            )
            profile_subtracted_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=profile_subtracted_image_1d
            )

            assert profile_subtracted_image_1d == pytest.approx(
                fit.profile_subtracted_image_1d
            )
            assert profile_subtracted_image_2d == pytest.approx(
                fit.profile_subtracted_image(return_in_2d=True)
            )

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=lens_data_7x7.noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
                mapper=mapper,
                regularization=reg,
            )

            model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d
            model_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=model_image_1d
            )

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=lens_data_7x7.image_1d,
                mask=lens_data_7x7.mask_1d,
                model_data=model_image_1d,
            )
            residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=residual_map_1d
            )

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

            normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=lens_data_7x7.noise_map_1d,
            )

            normalized_residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=normalized_residual_map_1d
            )

            assert normalized_residual_map_1d == pytest.approx(
                fit.normalized_residual_map_1d, 1e-4
            )
            assert normalized_residual_map_2d == pytest.approx(
                fit.normalized_residual_map(return_in_2d=True)
            )

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=lens_data_7x7.noise_map_1d,
            )
            chi_squared_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=chi_squared_map_1d
            )

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True)
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_7x7.mask_2d
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_7x7.mask_2d,
                noise_map=lens_data_7x7.noise_map(return_in_2d=True),
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4
            )

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___lens_fit_galaxy_image_dict__has_blurred_profile_images_and_inversion_reconstructed_data(
            self, lens_data_7x7
        ):

            g0 = g.Galaxy(
                redshift=0.5, light_profile=lp.EllipticalSersic(intensity=1.0)
            )
            g1 = g.Galaxy(
                redshift=0.5, light_profile=lp.EllipticalSersic(intensity=2.0)
            )
            g2 = g.Galaxy(redshift=0.5)

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[g0, g1, g2, galaxy_pix],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            g0_image_plane_image_1d = g0.intensities_from_grid(
                grid=lens_data_7x7.grid_stack.sub,
                return_in_2d=False,
                return_binned=True,
            )
            g0_image_plane_bluring_image_1d = g0.intensities_from_grid(
                grid=lens_data_7x7.grid_stack.blurring,
                return_in_2d=False,
                return_binned=False,
            )

            g0_blurred_image_plane_image_1d = lens_data_7x7.convolver_image.convolve_image(
                image_array=g0_image_plane_image_1d,
                blurring_array=g0_image_plane_bluring_image_1d,
            )

            g1_image_plane_image_1d = g1.intensities_from_grid(
                grid=tracer.source_plane.grid_stack.sub,
                return_in_2d=False,
                return_binned=True,
            )
            g1_image_plane_bluring_image_1d = g1.intensities_from_grid(
                grid=tracer.source_plane.grid_stack.blurring,
                return_in_2d=False,
                return_binned=False,
            )

            g1_blurred_image_plane_image_1d = lens_data_7x7.convolver_image.convolve_image(
                image_array=g1_image_plane_image_1d,
                blurring_array=g1_image_plane_bluring_image_1d,
            )

            blurred_profile_image_1d = (
                g0_blurred_image_plane_image_1d + g1_blurred_image_plane_image_1d
            )

            profile_subtracted_image_1d = (
                lens_data_7x7.image_1d - blurred_profile_image_1d
            )
            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=lens_data_7x7.noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
                mapper=mapper,
                regularization=reg,
            )

            assert fit.galaxy_image_1d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_1d, 1.0e-4
            )
            assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_1d, 1.0e-4
            )
            assert (fit.galaxy_image_1d_dict[g2] == np.zeros(9)).all()
            assert fit.galaxy_image_1d_dict[galaxy_pix] == pytest.approx(
                inversion.reconstructed_data_1d, 1.0e-4
            )

            assert fit.model_image_1d == pytest.approx(
                fit.galaxy_image_1d_dict[g0]
                + fit.galaxy_image_1d_dict[g1]
                + fit.galaxy_image_1d_dict[galaxy_pix],
                1.0e-4,
            )

            g0_blurred_image_plane_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=g0_blurred_image_plane_image_1d
            )
            g1_blurred_image_plane_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=g1_blurred_image_plane_image_1d
            )

            assert fit.galaxy_image_2d_dict[g0] == pytest.approx(
                g0_blurred_image_plane_image_2d, 1.0e-4
            )
            assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
                g1_blurred_image_plane_image_2d, 1.0e-4
            )
            assert fit.galaxy_image_2d_dict[galaxy_pix] == pytest.approx(
                inversion.reconstructed_data_2d, 1.0e-4
            )

            assert fit.model_image(return_in_2d=True) == pytest.approx(
                fit.galaxy_image_2d_dict[g0]
                + fit.galaxy_image_2d_dict[g1]
                + inversion.reconstructed_data_2d,
                1.0e-4,
            )

        def test___all_lens_fit_quantities__include_hyper_methods(self, lens_data_7x7):

            hyper_noise_map_max = 0.2
            lens_data_7x7.hyper_noise_map_max = hyper_noise_map_max

            hyper_image_sky = hi.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = hi.HyperBackgroundNoise(noise_scale=1.0)

            image_1d = hyper_image_sky.image_scaled_sky_from_image(
                image=lens_data_7x7.image_1d
            )

            hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
                noise_map=lens_data_7x7.noise_map_1d
            )

            galaxy_light = g.Galaxy(
                redshift=0.5,
                light_profile=lp.EllipticalSersic(intensity=1.0),
                hyper_galaxy=g.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image_1d=np.ones(9),
                hyper_galaxy_image_1d=np.ones(9),
                hyper_minimum_value=0.0,
            )

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[galaxy_light, galaxy_pix],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
                noise_map_1d=lens_data_7x7.noise_map_1d
            )
            hyper_noise_map_1d = hyper_noise_map_background_1d + hyper_noise_1d
            hyper_noise_map_1d[
                hyper_noise_map_1d > hyper_noise_map_max
            ] = hyper_noise_map_max

            hyper_noise_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=hyper_noise_map_1d
            )

            assert hyper_noise_map_1d == pytest.approx(fit.noise_map_1d, 1e-4)
            assert hyper_noise_map_2d == pytest.approx(fit.noise_map(return_in_2d=True))

            blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_7x7.convolver_image
            )

            blurred_profile_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=blurred_profile_image_1d
            )

            assert blurred_profile_image_1d == pytest.approx(
                fit.blurred_profile_image_1d, 1e-4
            )
            assert blurred_profile_image_2d == pytest.approx(
                fit.blurred_profile_image(return_in_2d=True)
            )

            profile_subtracted_image_1d = image_1d - blurred_profile_image_1d
            profile_subtracted_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=profile_subtracted_image_1d
            )

            assert profile_subtracted_image_1d == pytest.approx(
                fit.profile_subtracted_image_1d
            )
            assert profile_subtracted_image_2d == pytest.approx(
                fit.profile_subtracted_image(return_in_2d=True)
            )

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=hyper_noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
                mapper=mapper,
                regularization=reg,
            )

            model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d
            model_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=model_image_1d
            )

            assert model_image_1d == pytest.approx(fit.model_image_1d, 1e-4)
            assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

            residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
                data=image_1d, mask=lens_data_7x7.mask_1d, model_data=model_image_1d
            )

            residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=residual_map_1d
            )

            assert residual_map_1d == pytest.approx(fit.residual_map_1d, 1e-4)
            assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

            normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=hyper_noise_map_1d,
            )

            normalized_residual_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=normalized_residual_map_1d
            )

            assert normalized_residual_map_1d == pytest.approx(
                fit.normalized_residual_map_1d, 1e-4
            )
            assert normalized_residual_map_2d == pytest.approx(
                fit.normalized_residual_map(return_in_2d=True)
            )

            chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
                residual_map=residual_map_1d,
                mask=lens_data_7x7.mask_1d,
                noise_map=hyper_noise_map_1d,
            )

            chi_squared_map_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=chi_squared_map_1d
            )

            assert chi_squared_map_1d == pytest.approx(fit.chi_squared_map_1d, 1e-4)
            assert chi_squared_map_2d == pytest.approx(
                fit.chi_squared_map(return_in_2d=True)
            )

            chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
                chi_squared_map=chi_squared_map_2d, mask=lens_data_7x7.mask_2d
            )

            noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
                mask=lens_data_7x7.mask_2d, noise_map=hyper_noise_map_2d
            )

            likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert likelihood == pytest.approx(fit.likelihood, 1e-4)

            likelihood_with_regularization = util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert likelihood_with_regularization == pytest.approx(
                fit.likelihood_with_regularization, 1e-4
            )

            evidence = util.evidence_from_inversion_terms(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert evidence == fit.evidence
            assert evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
            self, lens_data_7x7
        ):
            galaxy_light = g.Galaxy(
                redshift=0.5, light_profile=lp.EllipticalSersic(intensity=1.0)
            )

            pix = pixelizations.Rectangular(shape=(3, 3))
            reg = regularization.Constant(coefficient=1.0)
            galaxy_pix = g.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
                galaxies=[galaxy_light, galaxy_pix],
                image_plane_grid_stack=lens_data_7x7.grid_stack,
                border=None,
            )

            fit = lens_fit.LensDataFit.for_data_and_tracer(
                lens_data=lens_data_7x7, tracer=tracer
            )

            blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=lens_data_7x7.convolver_image
            )

            blurred_profile_image_2d = lens_data_7x7.scaled_array_2d_from_array_1d(
                array_1d=blurred_profile_image_1d
            )

            profile_subtracted_image_1d = (
                lens_data_7x7.image_1d - blurred_profile_image_1d
            )

            mapper = pix.mapper_from_grid_stack_and_border(
                grid_stack=lens_data_7x7.grid_stack, border=None
            )

            inversion = inversions.Inversion.from_data_1d_mapper_and_regularization(
                image_1d=profile_subtracted_image_1d,
                noise_map_1d=lens_data_7x7.noise_map_1d,
                convolver=lens_data_7x7.convolver_mapping_matrix,
                mapper=mapper,
                regularization=reg,
            )

            assert blurred_profile_image_2d == pytest.approx(
                fit.model_image_2d_of_planes[0], 1.0e-4
            )
            assert inversion.reconstructed_data_2d == pytest.approx(
                fit.model_image_2d_of_planes[1], 1.0e-4
            )


class MockTracerPositions:
    def __init__(self, positions, noise=None):
        self.positions = positions
        self.noise = noise


class TestPositionFit:
    def test__x1_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(2)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(1.0) + np.square(3.0))

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [-9.0, -4.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_positions__mock_position_tracer__maximum_separation_is_correct(
        self
    ):
        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]])]
        )
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]])]
        )
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])]
        )
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(
            positions=[
                np.array(
                    [
                        [-2.0, -4.0],
                        [1.0, 3.0],
                        [0.1, 0.1],
                        [-0.1, -0.1],
                        [0.3, 0.4],
                        [-0.6, 0.5],
                    ]
                )
            ]
        )
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(
            positions=[np.array([[8.0, 4.0], [8.0, 4.0], [-9.0, -4.0]])]
        )
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_sets_of_positions__multiple_sets_of_max_distances(self):
        tracer = MockTracerPositions(
            positions=[
                np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]]),
            ]
        )

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separations[0] == 1.0
        assert fit.maximum_separations[1] == np.sqrt(18)
        assert fit.maximum_separations[2] == np.sqrt(18)

    def test__likelihood__is_sum_of_separations_divided_by_noise(self):
        tracer = MockTracerPositions(
            positions=[
                np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]]),
            ]
        )

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.chi_squared_map[0] == 1.0
        assert fit.chi_squared_map[1] == pytest.approx(18.0, 1e-4)
        assert fit.chi_squared_map[2] == pytest.approx(18.0, 1e-4)
        assert fit.figure_of_merit == pytest.approx(-0.5 * (1.0 + 18 + 18), 1e-4)

        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=2.0)
        assert fit.chi_squared_map[0] == (1.0 / 2.0) ** 2.0
        assert fit.chi_squared_map[1] == pytest.approx(
            (np.sqrt(18.0) / 2.0) ** 2.0, 1e-4
        )
        assert fit.chi_squared_map[2] == pytest.approx(
            (np.sqrt(18.0) / 2.0) ** 2.0, 1e-4
        )
        assert fit.figure_of_merit == pytest.approx(
            -0.5
            * (
                (1.0 / 2.0) ** 2.0
                + (np.sqrt(18.0) / 2.0) ** 2.0
                + (np.sqrt(18.0) / 2.0) ** 2.0
            ),
            1e-4,
        )

    def test__threshold__if_not_met_returns_ray_tracing_exception(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = lens_fit.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separation_within_threshold(threshold=100.0)
        assert not fit.maximum_separation_within_threshold(threshold=0.1)



from autolens.lens.util import lens_fit_util as util


class TestInversionEvidence:
    def test__simple_values(self):

        likelihood_with_regularization_terms = util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=3.0, regularization_term=6.0, noise_normalization=2.0
        )

        assert likelihood_with_regularization_terms == -0.5 * (3.0 + 6.0 + 2.0)

        evidences = util.evidence_from_inversion_terms(
            chi_squared=3.0,
            regularization_term=6.0,
            log_curvature_regularization_term=9.0,
            log_regularization_term=10.0,
            noise_normalization=30.0,
        )

        assert evidences == -0.5 * (3.0 + 6.0 + 9.0 - 10.0 + 30.0)



import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids, mask
from autolens.model.inversion.util import inversion_util


class TestDataVectorFromData(object):
    def test__simple_blurred_mapping_matrix__correct_data_vector(self):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        image = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        data_vector = inversion_util.data_vector_from_blurred_mapping_matrix_and_data(
            blurred_mapping_matrix=blurred_mapping_matrix,
            image_1d=image,
            noise_map_1d=noise_map,
        )

        assert (data_vector == np.array([2.0, 3.0, 1.0])).all()

    def test__simple_blurred_mapping_matrix__change_image_values__correct_data_vector(
        self
    ):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        image = np.array([3.0, 1.0, 1.0, 10.0, 1.0, 1.0])
        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        data_vector = inversion_util.data_vector_from_blurred_mapping_matrix_and_data(
            blurred_mapping_matrix=blurred_mapping_matrix,
            image_1d=image,
            noise_map_1d=noise_map,
        )

        assert (data_vector == np.array([4.0, 14.0, 10.0])).all()

    def test__simple_blurred_mapping_matrix__change_noise_values__correct_data_vector(
        self
    ):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        image = np.array([4.0, 1.0, 1.0, 16.0, 1.0, 1.0])
        noise_map = np.array([2.0, 1.0, 1.0, 4.0, 1.0, 1.0])

        data_vector = inversion_util.data_vector_from_blurred_mapping_matrix_and_data(
            blurred_mapping_matrix=blurred_mapping_matrix,
            image_1d=image,
            noise_map_1d=noise_map,
        )

        assert (data_vector == np.array([2.0, 3.0, 1.0])).all()


class TestCurvatureMatrixFromBlurred(object):
    def test__simple_blurred_mapping_matrix(self):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        noise_map = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        curvature_matrix = inversion_util.curvature_matrix_from_blurred_mapping_matrix(
            blurred_mapping_matrix=blurred_mapping_matrix, noise_map_1d=noise_map
        )

        assert (
            curvature_matrix
            == np.array([[2.0, 1.0, 0.0], [1.0, 3.0, 1.0], [0.0, 1.0, 1.0]])
        ).all()

    def test__simple_blurred_mapping_matrix__change_noise_values(self):

        blurred_mapping_matrix = np.array(
            [
                [1.0, 1.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        noise_map = np.array([2.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        curvature_matrix = inversion_util.curvature_matrix_from_blurred_mapping_matrix(
            blurred_mapping_matrix=blurred_mapping_matrix, noise_map_1d=noise_map
        )

        assert (
            curvature_matrix
            == np.array([[1.25, 0.25, 0.0], [0.25, 2.25, 1.0], [0.0, 1.0, 1.0]])
        ).all()


class TestPixelizationResiduals(object):
    def test__pixelization_perfectly_reconstructed_data__quantities_like_residuals_all_zeros(
        self
    ):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.ones(9)
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]

        pixelization_residuals = inversion_util.pixelization_residual_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_residuals == np.zeros(3)).all()

    def test__pixelization_not_perfect_fit__quantities_like_residuals_non_zero(self):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = 2.0 * np.ones(9)
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_residuals = inversion_util.pixelization_residual_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_residuals == 3.0 * np.ones(3)).all()

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.array([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_residuals = inversion_util.pixelization_residual_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_residuals == np.array([0.0, 3.0, 6.0])).all()


class TestPixelizationNormalizedResiduals(object):
    def test__pixelization_perfectly_reconstructed_data__quantities_like_residuals_all_zeros(
        self
    ):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.ones(9)
        noise_map_1d = np.ones(9)
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]

        pixelization_normalized_residuals = inversion_util.pixelization_normalized_residual_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_normalized_residuals == np.zeros(3)).all()

    def test__pixelization_not_perfect_fit__quantities_like_residuals_non_zero(self):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = 2.0 * np.ones(9)
        noise_map_1d = 2.0 * np.ones(9)
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_normalized_residuals = inversion_util.pixelization_normalized_residual_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_normalized_residuals == 1.5 * np.ones(3)).all()

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.array([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
        noise_map_1d = np.array([0.5, 0.5, 0.5, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0])
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_normalized_residuals = inversion_util.pixelization_normalized_residual_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_normalized_residuals == np.array([0.0, 3.0, 3.0])).all()


class TestPixelizationChiSquareds(object):
    def test__pixelization_perfectly_reconstructed_data__quantities_like_residuals_all_zeros(
        self
    ):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.ones(9)
        noise_map_1d = np.ones(9)
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]

        pixelization_chi_squareds = inversion_util.pixelization_chi_squared_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_chi_squareds == np.zeros(3)).all()

    def test__pixelization_not_perfect_fit__quantities_like_residuals_non_zero(self):

        pixelization_values = np.ones(3)
        reconstructed_data_1d = 2.0 * np.ones(9)
        noise_map_1d = 2.0 * np.ones(9)
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_chi_squareds = inversion_util.pixelization_chi_squared_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_chi_squareds == 0.75 * np.ones(3)).all()

        pixelization_values = np.ones(3)
        reconstructed_data_1d = np.array([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])
        noise_map_1d = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 4.0, 4.0, 4.0])
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        pixelization_to_sub_all = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        pixelization_chi_squareds = inversion_util.pixelization_chi_squared_map_from_pixelization_values_and_reconstructed_data_1d(
            pixelization_values=pixelization_values,
            reconstructed_data_1d=reconstructed_data_1d,
            noise_map_1d=noise_map_1d,
            sub_to_regular=sub_to_regular,
            pixelization_to_sub_all=pixelization_to_sub_all,
        )

        assert (pixelization_chi_squareds == np.array([0.0, 12.0, 0.75])).all()



from autofit.tools import path_util
from autofit import conf
from autolens.data.array.util import array_util
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.plotters import ccd_plotters

import os
import numpy as np

slacs = []

slacs.append(
    {
        "name": "slacs0216-0813",
        "target_name": "GAL-0668-52162-428",
        "exposure_time": 2232.0,
        "redshift_lens": 0.332,
        "redshift_source": 0.523,
    }
)
slacs.append(
    {
        "name": "slacs0252+0039",
        "target_name": "GAL-0807-52295-614",
        "exposure_time": 2058.0,
        "redshift_lens": 0.280,
        "redshift_source": 0.982,
    }
)
slacs.append(
    {
        "name": "slacs0737+3216",
        "target_name": "GAL-0541-51959-145",
        "exposure_time": 2272.0,
        "redshift_lens": 0.322,
        "redshift_source": 0.581,
    }
)
slacs.append(
    {
        "name": "slacs0912+0029",
        "target_name": "GAL-0472-51955-429",
        "exposure_time": 2224.0,
        "redshift_lens": 0.164,
        "redshift_source": 0.324,
    }
)  # two exposures - was highest time used?
slacs.append(
    {
        "name": "slacs0959+4410",
        "target_name": "GAL-0572-52289-495",
        "exposure_time": 2224.0,
        "redshift_lens": 0.126,
        "redshift_source": 0.535,
    }
)
slacs.append(
    {
        "name": "slacs0959+4416",
        "target_name": "GAL-0942-52703-499",
        "exposure_time": 2128.0,
        "redshift_lens": 0.237,
        "redshift_source": 0.531,
    }
)
slacs.append(
    {
        "name": "slacs0959+5100",
        "target_name": "GAL-0902-52409-068",
        "exposure_time": 2440.0,
        "redshift_lens": 0.241,
        "redshift_source": 0.470,
    }
)
slacs.append(
    {
        "name": "slacs1011+0143",
        "target_name": "GAL-PLACE_HOLDER",
        "exposure_time": 2088.0,
        "redshift_lens": 0.331,
        "redshift_source": 2.701,
    }
)
slacs.append(
    {
        "name": "slacs1205+4910",
        "target_name": "GAL-0969-52442-134",
        "exposure_time": 2388.0,
        "redshift_lens": 0.215,
        "redshift_source": 0.481,
    }
)
slacs.append(
    {
        "name": "slacs1250+0523",
        "target_name": "GAL-0847-52426-549",
        "exposure_time": 2232.0,
        "redshift_lens": 0.232,
        "redshift_source": 0.795,
    }
)
slacs.append(
    {
        "name": "slacs1402+6321",
        "target_name": "GAL-0605-52353-503",
        "exposure_time": 2520.0,
        "redshift_lens": 0.205,
        "redshift_source": 0.481,
    }
)
slacs.append(
    {
        "name": "slacs1420+6019",
        "target_name": "GAL-0788-52338-605",
        "exposure_time": 2520.0,
        "redshift_lens": 0.063,
        "redshift_source": 0.535,
    }
)
slacs.append(
    {
        "name": "slacs1430+4105",
        "target_name": "GAL-1349-52797-406",
        "exposure_time": 2128.0,
        "redshift_lens": 0.285,
        "redshift_source": 0.575,
    }
)
slacs.append(
    {
        "name": "slacs1627+0053",
        "target_name": "GAL-0364-52000-084",
        "exposure_time": 2240.0,
        "redshift_lens": 0.208,
        "redshift_source": 0.524,
    }
)
slacs.append(
    {
        "name": "slacs1630+4520",
        "target_name": "GAL-0626-52057-518",
        "exposure_time": 2388.0,
        "redshift_lens": 0.248,
        "redshift_source": 0.793,
    }
)
slacs.append(
    {
        "name": "slacs2238-0754",
        "target_name": "GAL-0722-52224-442",
        "exposure_time": 2232.0,
        "redshift_lens": 0.137,
        "redshift_source": 0.713,
    }
)
slacs.append(
    {
        "name": "slacs2300+0022",
        "target_name": "GAL-0677-52606-520",
        "exposure_time": 2224.0,
        "redshift_lens": 0.228,
        "redshift_source": 0.463,
    }
)
slacs.append(
    {
        "name": "slacs2303+1422",
        "target_name": "GAL-0743-52262-304",
        "exposure_time": 2240.0,
        "redshift_lens": 0.155,
        "redshift_source": 0.517,
    }
)

workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
data_raw_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data_raw"]
)
slacs_sample_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=data_raw_path, folder_names=["slacs_sample"]
)
for slac in slacs:

    name = slac["name"]
    target_name = slac["target_name"]
    exposure_time = slac["exposure_time"]

    for filter in ["F814W", "F555W"]:

        slacs_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
            path=slacs_sample_path, folder_names=["david_0"]
        )
        slacs_file = slacs_data_path + target_name + "_" + filter + "_post.fits"

        if os.path.exists(slacs_file):

            ccd_data = ccd.load_ccd_data_from_fits(
                image_path=slacs_file,
                image_hdu=1,
                pixel_scale=0.03,
                psf_path=slacs_file,
                psf_hdu=3,
                noise_map_from_image_and_background_noise_map=True,
                background_noise_map_path=slacs_file,
                background_noise_map_hdu=2,
                convert_background_noise_map_from_inverse_noise_map=True,
                poisson_noise_map_from_image=True,
                exposure_time_map_from_single_value=exposure_time,
                exposure_time_map_from_inverse_noise_map=True,
                resized_psf_shape=(61, 61),
            )

            ccd_data.noise_map = array_util.replace_noise_map_2d_values_where_image_2d_values_are_negative(
                image_2d=ccd_data.image, noise_map_2d=ccd_data.noise_map
            )

            print()
            print(
                name + "  " + filter + "  " + str(ccd_data.absolute_signal_to_noise_max)
            )

            slacs_output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=workspace_path, folder_names=["data", "slacs", name]
            )

            ccd.output_ccd_data_to_fits(
                ccd_data=ccd_data,
                image_path=slacs_output_path + filter + "_image.fits",
                psf_path=slacs_output_path + filter + "_psf.fits",
                noise_map_path=slacs_output_path + filter + "_noise_map.fits",
                background_noise_map_path=slacs_output_path
                + filter
                + "_background_noise_map.fits",
                poisson_noise_map_path=slacs_output_path
                + filter
                + "_poisson_noise_map.fits",
                exposure_time_map_path=slacs_output_path
                + filter
                + "_exposure_time_map.fits",
                overwrite=True,
            )

            slacs_image_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=slacs_output_path, folder_names=["images"]
            )

            ccd_plotters.plot_ccd_subplot(
                ccd_data=ccd_data,
                output_path=slacs_image_path + filter + "_",
                output_format="png",
            )
            ccd_plotters.plot_ccd_individual(
                ccd_data=ccd_data,
                should_plot_image=True,
                should_plot_noise_map=True,
                should_plot_psf=True,
                should_plot_signal_to_noise_map=True,
                should_plot_absolute_signal_to_noise_map=True,
                should_plot_potential_chi_squared_map=True,
                output_path=slacs_image_path + filter + "_",
                output_format="png",
            )



import numpy as np
from astropy import cosmology as cosmo
from skimage import measure

import autofit as af
from autolens import exc, dimensions as dim
from autolens.data.array import scaled_array
from autolens.data.array.grids import (
    reshape_returned_array,
    reshape_returned_array_blurring,
    reshape_returned_grid,
)
from autolens.lens.util import lens_util
from autolens.model import cosmology_util
from autolens.data.array.util import grid_util


class AbstractPlane(object):
    def __init__(self, redshift, galaxies=None, cosmology=cosmo.Planck15):
        """An abstract plane which represents a set of galaxies that are close to one another in redshift-space.

        From an abstract plane, cosmological quantities like angular diameter distances can be computed. If the  \
        redshift of the plane is input as *None*, quantities that depend on a redshift are returned as *None*.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of galaxies in this plane.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        self.redshift = redshift
        self.galaxies = galaxies
        self.cosmology = cosmology

    @property
    def galaxy_redshifts(self):
        return [galaxy.redshift for galaxy in self.galaxies]

    @property
    def arcsec_per_kpc(self):
        return cosmology_util.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology
        )

    @property
    def kpc_per_arcsec(self):
        return 1.0 / self.arcsec_per_kpc

    def angular_diameter_distance_to_earth_in_units(self, unit_length="arcsec"):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.redshift, cosmology=self.cosmology, unit_length=unit_length
        )

    def cosmic_average_density_in_units(
        self, unit_length="arcsec", unit_mass="solMass"
    ):
        return cosmology_util.cosmic_average_density_from_redshift_and_cosmology(
            redshift=self.redshift,
            cosmology=self.cosmology,
            unit_length=unit_length,
            unit_mass=unit_mass,
        )

    @property
    def has_light_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_light_profile, self.galaxies)))

    @property
    def has_mass_profile(self):
        return any(list(map(lambda galaxy: galaxy.has_mass_profile, self.galaxies)))

    @property
    def has_pixelization(self):
        return any([galaxy.pixelization for galaxy in self.galaxies])

    @property
    def has_regularization(self):
        return any([galaxy.regularization for galaxy in self.galaxies])

    @property
    def regularization(self):

        galaxies_with_regularization = list(
            filter(lambda galaxy: galaxy.has_regularization, self.galaxies)
        )

        if len(galaxies_with_regularization) == 0:
            return None
        if len(galaxies_with_regularization) == 1:
            return galaxies_with_regularization[0].regularization
        elif len(galaxies_with_regularization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with regularizations in one plane is above 1"
            )

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda galaxy: galaxy.has_hyper_galaxy, self.galaxies)))

    @property
    def contribution_maps_1d_of_galaxies(self):

        contribution_maps_1d = []

        for galaxy in self.galaxies:

            if galaxy.hyper_galaxy is not None:

                contribution_map = galaxy.hyper_galaxy.contribution_map_from_hyper_images(
                    hyper_model_image=galaxy.hyper_model_image_1d,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image_1d,
                )

                contribution_maps_1d.append(contribution_map)

            else:

                contribution_maps_1d.append(None)

        return contribution_maps_1d

    @property
    def centres_of_galaxy_mass_profiles(self):

        galaxies_with_mass_profiles = [
            galaxy for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

        mass_profile_centres = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_centres[galaxy_index] = [
                profile.centre for profile in galaxy.mass_profiles
            ]
        return mass_profile_centres

    @property
    def axis_ratios_of_galaxy_mass_profiles(self):
        galaxies_with_mass_profiles = [
            galaxy for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

        mass_profile_axis_ratios = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_axis_ratios[galaxy_index] = [
                profile.axis_ratio for profile in galaxy.mass_profiles
            ]
        return mass_profile_axis_ratios

    @property
    def phis_of_galaxy_mass_profiles(self):

        galaxies_with_mass_profiles = [
            galaxy for galaxy in self.galaxies if galaxy.has_mass_profile
        ]

        mass_profile_phis = [[] for _ in range(len(galaxies_with_mass_profiles))]

        for galaxy_index, galaxy in enumerate(galaxies_with_mass_profiles):
            mass_profile_phis[galaxy_index] = [
                profile.phi for profile in galaxy.mass_profiles
            ]
        return mass_profile_phis

    def luminosities_of_galaxies_within_circles_in_units(
        self, radius: dim.Length, unit_luminosity="eps", exposure_time=None
    ):
        """Compute the total luminosity of all galaxies in this plane within a circle of specified radius.

        See *galaxy.light_within_circle* and *light_profiles.light_within_circle* for details \
        of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        return list(
            map(
                lambda galaxy: galaxy.luminosity_within_circle_in_units(
                    radius=radius,
                    unit_luminosity=unit_luminosity,
                    exposure_time=exposure_time,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def luminosities_of_galaxies_within_ellipses_in_units(
        self, major_axis: dim.Length, unit_luminosity="eps", exposure_time=None
    ):
        """
        Compute the total luminosity of all galaxies in this plane within a ellipse of specified major-axis.

        The value returned by this integral is dimensionless, and a conversion factor can be specified to convert it \
        to a physical value (e.g. the photometric zeropoint).

        See *galaxy.light_within_ellipse* and *light_profiles.light_within_ellipse* for details
        of how this is performed.

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        return list(
            map(
                lambda galaxy: galaxy.luminosity_within_ellipse_in_units(
                    major_axis=major_axis,
                    unit_luminosity=unit_luminosity,
                    exposure_time=exposure_time,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def masses_of_galaxies_within_circles_in_units(
        self, radius: dim.Length, unit_mass="solMass", redshift_source=None
    ):
        """Compute the total mass of all galaxies in this plane within a circle of specified radius.

        See *galaxy.angular_mass_within_circle* and *mass_profiles.angular_mass_within_circle* for details
        of how this is performed.

        Parameters
        ----------
        redshift_source
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The units the mass is returned in (angular | solMass).

        """
        return list(
            map(
                lambda galaxy: galaxy.mass_within_circle_in_units(
                    radius=radius,
                    unit_mass=unit_mass,
                    redshift_source=redshift_source,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def masses_of_galaxies_within_ellipses_in_units(
        self, major_axis: dim.Length, unit_mass="solMass", redshift_source=None
    ):
        """Compute the total mass of all galaxies in this plane within a ellipse of specified major-axis.

        See *galaxy.angular_mass_within_ellipse* and *mass_profiles.angular_mass_within_ellipse* for details \
        of how this is performed.

        Parameters
        ----------
        redshift_source
        unit_mass
        major_axis : float
            The major-axis radius of the ellipse.

        """
        return list(
            map(
                lambda galaxy: galaxy.mass_within_ellipse_in_units(
                    major_axis=major_axis,
                    unit_mass=unit_mass,
                    redshift_source=redshift_source,
                    cosmology=self.cosmology,
                ),
                self.galaxies,
            )
        )

    def einstein_radius_in_units(self, unit_length="arcsec"):

        if self.has_mass_profile:
            return sum(
                filter(
                    None,
                    list(
                        map(
                            lambda galaxy: galaxy.einstein_radius_in_units(
                                unit_length=unit_length, cosmology=self.cosmology
                            ),
                            self.galaxies,
                        )
                    ),
                )
            )

    def einstein_mass_in_units(self, unit_mass="solMass", redshift_source=None):

        if self.has_mass_profile:
            return sum(
                filter(
                    None,
                    list(
                        map(
                            lambda galaxy: galaxy.einstein_mass_in_units(
                                unit_mass=unit_mass,
                                redshift_source=redshift_source,
                                cosmology=self.cosmology,
                            ),
                            self.galaxies,
                        )
                    ),
                )
            )

    # noinspection PyUnusedLocal
    def summarize_in_units(
        self,
        radii,
        whitespace=80,
        unit_length="arcsec",
        unit_luminosity="eps",
        unit_mass="solMass",
        redshift_source=None,
        **kwargs
    ):

        summary = ["Plane\n"]
        prefix_plane = ""

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_plane + "redshift",
                value=self.redshift,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_plane + "kpc_per_arcsec",
                value=self.kpc_per_arcsec,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        angular_diameter_distance_to_earth = self.angular_diameter_distance_to_earth_in_units(
            unit_length=unit_length
        )

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_plane + "angular_diameter_distance_to_earth",
                value=angular_diameter_distance_to_earth,
                whitespace=whitespace,
                format_string="{:.2f}",
            )
        ]

        for galaxy in self.galaxies:
            summary += ["\n"]
            summary += galaxy.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                unit_mass=unit_mass,
                redshift_source=redshift_source,
                cosmology=self.cosmology,
            )

        return summary


class AbstractGriddedPlane(AbstractPlane):
    def __init__(
        self,
        redshift,
        grid_stack,
        border,
        compute_deflections,
        galaxies=None,
        cosmology=cosmo.Planck15,
    ):
        """An abstract plane which represents a set of galaxies that are close to one another in redshift-space and \
        have an associated grid on which lensing calcuations are performed.

        From an abstract plane grid, the surface-density, potential and deflection angles of the galaxies can be \
        computed.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        grid_stack : masks.GridStack
            The stack of grid_stacks of (y,x) arc-second coordinates of this plane.
        border : masks.RegularGridBorder
            The borders of the regular-grid, which is used to relocate demagnified traced regular-pixel to the \
            source-plane borders.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        super(AbstractGriddedPlane, self).__init__(
            redshift=redshift, galaxies=galaxies, cosmology=cosmology
        )

        self.grid_stack = grid_stack
        self.border = border

        if compute_deflections:

            def calculate_deflections(grid):

                if galaxies:
                    return sum(
                        map(
                            lambda galaxy: galaxy.deflections_from_grid(
                                grid, return_in_2d=False, return_binned=False
                            ),
                            galaxies,
                        )
                    )
                else:
                    return np.full((grid.shape[0], 2), 0.0)

            self.deflections_stack = grid_stack.apply_function(calculate_deflections)

        else:

            self.deflections_stack = None

    def trace_grid_stack_to_next_plane(self):
        """Trace this plane's grid_stacks to the next plane, using its deflection angles."""

        def minus(grid, deflections):
            return grid - deflections

        return self.grid_stack.map_function(minus, self.deflections_stack)

    @reshape_returned_array
    def profile_image_plane_image(self, return_in_2d=True, return_binned=True):
        """Compute the profile-image plane image of the list of galaxies of the plane's sub-grid, by summing the
        individual images of each galaxy's light profile.

        The image is calculated on the sub-grid and binned-up to the original regular grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the regular grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.intensities_from_grid(
                        grid=self.grid_stack.sub,
                        return_in_2d=False,
                        return_binned=False,
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((self.grid_stack.sub.shape[0]), 0.0)

    def profile_image_plane_image_of_galaxies(
        self, return_in_2d=True, return_binned=True
    ):
        return list(
            map(
                lambda galaxy: self.profile_image_plane_image_of_galaxy(
                    galaxy=galaxy,
                    return_in_2d=return_in_2d,
                    return_binned=return_binned,
                ),
                self.galaxies,
            )
        )

    def profile_image_plane_image_of_galaxy(
        self, galaxy, return_in_2d=True, return_binned=True
    ):
        return galaxy.intensities_from_grid(
            grid=self.grid_stack.sub,
            return_in_2d=return_in_2d,
            return_binned=return_binned,
        )

    @reshape_returned_array_blurring
    def profile_image_plane_blurring_image(self, return_in_2d=True):
        """Compute the profile-image plane blurring image of the list of galaxies of the plane's sub-grid, by summing \
        the individual blurring images of each galaxy's light profile.

        The blurring image is calculated on the blurring grid, which is not subgrided.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        blurring grid is returned.

        Parameters
        -----------
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.intensities_from_grid(
                        grid=self.grid_stack.blurring,
                        return_in_2d=False,
                        return_binned=False,
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((self.grid_stack.bluring.shape[0]), 0.0)

    def profile_image_plane_blurring_image_of_galaxies(self, return_in_2d=False):
        return list(
            map(
                lambda galaxy: self.profile_image_plane_blurring_image_of_galaxy(
                    galaxy=galaxy, return_in_2d=return_in_2d
                ),
                self.galaxies,
            )
        )

    def profile_image_plane_blurring_image_of_galaxy(self, galaxy, return_in_2d=False):
        return galaxy.intensities_from_grid(
            grid=self.grid_stack.blurring,
            return_in_2d=return_in_2d,
            return_binned=False,
        )

    @reshape_returned_array
    def convergence(self, return_in_2d=True, return_binned=True):
        """Compute the convergence of the list of galaxies of the plane's sub-grid, by summing the individual convergences \
        of each galaxy's mass profile.

        The convergence is calculated on the sub-grid and binned-up to the original regular grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        grid : RegularGrid
            The grid (regular or sub) of (y,x) arc-second coordinates at the centre of every unmasked pixel which the \
            potential is calculated on.
        galaxies : [galaxy.Galaxy]
            The galaxies whose mass profiles are used to compute the surface densities.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the regular grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.convergence_from_grid(
                        grid=self.grid_stack.sub.unlensed_grid_1d,
                        return_in_2d=False,
                        return_binned=False,
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((self.grid_stack.sub.shape[0]), 0.0)

    @reshape_returned_array
    def potential(self, return_in_2d=True, return_binned=True):
        """Compute the potential of the list of galaxies of the plane's sub-grid, by summing the individual potentials \
        of each galaxy's mass profile.

        The potential is calculated on the sub-grid and binned-up to the original regular grid by taking the mean
        value of every set of sub-pixels, provided the *returned_binned_sub_grid* bool is *True*.

        If the plane has no galaxies (or no galaxies have mass profiles) an array of all zeros the shape of the plane's
        sub-grid is returned.

        Parameters
        -----------
        grid : RegularGrid
            The grid (regular or sub) of (y,x) arc-second coordinates at the centre of every unmasked pixel which the \
            potential is calculated on.
        galaxies : [galaxy.Galaxy]
            The galaxies whose mass profiles are used to compute the surface densities.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the regular grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.potential_from_grid(
                        grid=self.grid_stack.sub.unlensed_grid_1d,
                        return_in_2d=False,
                        return_binned=False,
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((self.grid_stack.sub.shape[0]), 0.0)

    @reshape_returned_array
    def deflections_y(self, return_in_2d=True, return_binned=True):
        return self.deflections(return_in_2d=False, return_binned=False)[:, 0]

    @reshape_returned_array
    def deflections_x(self, return_in_2d=True, return_binned=True):
        return self.deflections(return_in_2d=False, return_binned=False)[:, 1]

    @reshape_returned_grid
    def deflections(self, return_in_2d=True, return_binned=True):
        if self.galaxies:
            return sum(
                map(
                    lambda g: g.deflections_from_grid(
                        grid=self.grid_stack.sub.unlensed_grid_1d,
                        return_in_2d=False,
                        return_binned=False,
                    ),
                    self.galaxies,
                )
            )
        else:
            return np.full((self.grid_stack.sub.shape[0], 2), 0.0)

    @property
    def plane_image(self):
        return lens_util.plane_image_of_galaxies_from_grid(
            shape=self.grid_stack.regular.mask.shape,
            grid=self.grid_stack.regular,
            galaxies=self.galaxies,
        )

    @property
    def mapper(self):

        galaxies_with_pixelization = list(
            filter(lambda galaxy: galaxy.pixelization is not None, self.galaxies)
        )

        if len(galaxies_with_pixelization) == 0:
            return None
        if len(galaxies_with_pixelization) == 1:

            pixelization = galaxies_with_pixelization[0].pixelization

            return pixelization.mapper_from_grid_stack_and_border(
                grid_stack=self.grid_stack,
                border=self.border,
                hyper_image=galaxies_with_pixelization[0].hyper_galaxy_image_1d,
            )

        elif len(galaxies_with_pixelization) > 1:
            raise exc.PixelizationException(
                "The number of galaxies with pixelizations in one plane is above 1"
            )

    @property
    def yticks(self):
        """Compute the yticks labels of this grid_stack, used for plotting the y-axis ticks when visualizing an image \
        """
        return np.linspace(
            np.amin(self.grid_stack.regular[:, 0]),
            np.amax(self.grid_stack.regular[:, 0]),
            4,
        )

    @property
    def xticks(self):
        """Compute the xticks labels of this grid_stack, used for plotting the x-axis ticks when visualizing an \
        image"""
        return np.linspace(
            np.amin(self.grid_stack.regular[:, 1]),
            np.amax(self.grid_stack.regular[:, 1]),
            4,
        )

    @reshape_returned_grid
    def deflections_via_potential(self, return_in_2d=True, return_binned=True):
        potential_2d = self.potential(return_in_2d=True, return_binned=False)

        deflections_y_2d = np.gradient(
            potential_2d, self.grid_stack.sub.in_2d[:, 0, 0], axis=0
        )
        deflections_x_2d = np.gradient(
            potential_2d, self.grid_stack.sub.in_2d[0, :, 1], axis=1
        )

        return np.stack((deflections_y_2d, deflections_x_2d), axis=-1)

    @reshape_returned_array
    def lensing_jacobian_a11_from_deflections_2d(
        self, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1.0 - np.gradient(
            deflections_2d[:, :, 1], self.grid_stack.in_2d[0, :, 1], axis=1
        )

    @reshape_returned_array
    def lensing_jacobian_a12_from_deflections_2d(
        self, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(
            deflections_2d[:, :, 1], self.grid_stack.in_2d[:, 0, 0], axis=0
        )

    @reshape_returned_array
    def lensing_jacobian_a21_from_deflections_2d(
        self, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(
            deflections_2d[:, :, 0], self.grid_stack.in_2d[0, :, 1], axis=1
        )

    @reshape_returned_array
    def lensing_jacobian_a22_from_deflections_2d(
        self, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1 - np.gradient(
            deflections_2d[:, :, 0], self.grid_stack.in_2d[:, 0, 0], axis=0
        )

    def lensing_jacobian(self, return_in_2d=True, return_binned=True):

        a11 = self.lensing_jacobian_a11_from_deflections_2d(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a12 = self.lensing_jacobian_a12_from_deflections_2d(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a21 = self.lensing_jacobian_a21_from_deflections_2d(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a22 = self.lensing_jacobian_a22_from_deflections_2d(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        return np.array([[a11, a12], [a21, a22]])

    @reshape_returned_array
    def convergence_from_jacobian(self, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian(return_in_2d=False, return_binned=False)

        convergence = 1 - 0.5 * (jacobian[0, 0] + jacobian[1, 1])

        return convergence

    @reshape_returned_array
    def shear_from_jacobian(self, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian(return_in_2d=True, return_binned=False)

        gamma_1 = 0.5 * (jacobian[1, 1] - jacobian[0, 0])
        gamma_2 = -0.5 * (jacobian[0, 1] + jacobian[1, 0])

        return (gamma_1 ** 2 + gamma_2 ** 2) ** 0.5

    @reshape_returned_array
    def tangential_eigen_value_from_shear_and_convergence(
        self, return_in_2d=True, return_binned=True
    ):

        convergence = self.convergence_from_jacobian(
            return_in_2d=False, return_binned=False
        )

        shear = self.shear_from_jacobian(return_in_2d=False, return_binned=False)

        return 1 - convergence - shear

    @reshape_returned_array
    def radial_eigen_value_from_shear_and_convergence(
        self, return_in_2d=True, return_binned=True
    ):

        convergence = self.convergence_from_jacobian(
            return_in_2d=False, return_binned=False
        )

        shear = self.shear_from_jacobian(return_in_2d=False, return_binned=False)

        return 1 - convergence + shear

    @reshape_returned_array
    def magnification_from_grid(self, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian(return_in_2d=False, return_binned=False)

        det_jacobian = jacobian[0, 0] * jacobian[1, 1] - jacobian[0, 1] * jacobian[1, 0]

        return 1 / det_jacobian

    def tangential_critical_curve_from_grid(self):

        lambda_tangential_2d = self.tangential_eigen_value_from_shear_and_convergence(
            return_in_2d=True, return_binned=False
        )

        tangential_critical_curve_indices = measure.find_contours(
            lambda_tangential_2d, 0
        )

        if tangential_critical_curve_indices == []:
            return []

        tangential_critical_curve = grid_util.grid_pixels_1d_to_grid_arcsec_1d(
            grid_pixels_1d=tangential_critical_curve_indices[0],
            shape=lambda_tangential_2d.shape,
            pixel_scales=(
                self.grid_stack.pixel_scale / self.grid_stack.sub_grid_size,
                self.grid_stack.pixel_scale / self.grid_stack.sub_grid_size,
            ),
            origin=self.grid_stack.mask.origin,
        )

        # Bug with offset, this fixes it for now

        tangential_critical_curve[:, 0] -= self.grid_stack.pixel_scale / 2.0
        tangential_critical_curve[:, 1] += self.grid_stack.pixel_scale / 2.0

        return tangential_critical_curve

    def tangential_caustic_from_grid(self):

        tangential_critical_curve = self.tangential_critical_curve_from_grid()

        if tangential_critical_curve == []:
            return []

        deflections_1d = self.deflections(
            grid=tangential_critical_curve, return_in_2d=False, return_binned=False
        )

        return tangential_critical_curve - deflections_1d

    def radial_critical_curve_from_grid(self, grid):

        lambda_radial_2d = self.radial_eigen_value_from_shear_and_convergence(
            grid=grid, return_in_2d=True, return_binned=False
        )

        radial_critical_curve_indices = measure.find_contours(lambda_radial_2d, 0)

        if radial_critical_curve_indices == []:
            return []

        radial_critical_curve = grid_util.grid_pixels_1d_to_grid_arcsec_1d(
            grid_pixels_1d=radial_critical_curve_indices[0],
            shape=lambda_radial_2d.shape,
            pixel_scales=(
                grid.pixel_scale / grid.sub_grid_size,
                grid.pixel_scale / grid.sub_grid_size,
            ),
            origin=grid.mask.origin,
        )

        # Bug with offset, this fixes it for now

        radial_critical_curve[:, 0] -= grid.pixel_scale / 2.0
        radial_critical_curve[:, 1] += grid.pixel_scale / 2.0

        return radial_critical_curve

    def radial_caustic_from_grid(self, grid):

        radial_critical_curve = self.radial_critical_curve_from_grid(grid=grid)

        if radial_critical_curve == []:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=radial_critical_curve, return_in_2d=False, return_binned=False
        )

        return radial_critical_curve - deflections_1d

    def critical_curves_from_grid(self, grid):
        return [
            self.tangential_critical_curve_from_grid(grid=grid),
            self.radial_critical_curve_from_grid(grid=grid),
        ]

    def caustics_from_grid(self, grid):
        return [
            self.tangential_caustic_from_grid(grid=grid),
            self.radial_caustic_from_grid(grid=grid),
        ]


class AbstractDataPlane(AbstractGriddedPlane):
    def blurred_profile_image_plane_image_1d_from_convolver_image(
        self, convolver_image
    ):

        image_array = self.profile_image_plane_image(
            return_in_2d=False, return_binned=True
        )
        blurring_array = self.profile_image_plane_blurring_image(return_in_2d=False)

        return convolver_image.convolve_image(
            image_array=image_array, blurring_array=blurring_array
        )

    def blurred_profile_image_plane_images_1d_of_galaxies_from_convolver_image(
        self, convolver_image
    ):

        return list(
            map(
                lambda profile_image_plane_image_1d, profile_image_plane_blurring_image_1d: convolver_image.convolve_image(
                    image_array=profile_image_plane_image_1d,
                    blurring_array=profile_image_plane_blurring_image_1d,
                ),
                self.profile_image_plane_image_of_galaxies(
                    return_in_2d=False, return_binned=True
                ),
                self.profile_image_plane_blurring_image_of_galaxies(return_in_2d=False),
            )
        )

    def visibilities_from_transformer(self, transformer):

        profile_image_plane_image_1d = self.profile_image_plane_image(
            return_in_2d=False, return_binned=True
        )

        return transformer.visibilities_from_intensities(
            intensities_1d=profile_image_plane_image_1d
        )

    def hyper_noise_map_1d_from_noise_map_1d(self, noise_map_1d):
        hyper_noise_maps_1d = self.hyper_noise_maps_1d_of_galaxies_from_noise_map_1d(
            noise_map_1d=noise_map_1d
        )
        hyper_noise_maps_1d = [
            hyper_noise_map
            for hyper_noise_map in hyper_noise_maps_1d
            if hyper_noise_map is not None
        ]
        return sum(hyper_noise_maps_1d)

    def hyper_noise_maps_1d_of_galaxies_from_noise_map_1d(self, noise_map_1d):
        """For a contribution map and noise-map, use the model hyper_galaxy galaxies to compute a hyper noise-map.

        Parameters
        -----------
        noise_map_1d : ccd.NoiseMap or ndarray
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        """
        hyper_noise_maps_1d = []

        for galaxy in self.galaxies:
            if galaxy.hyper_galaxy is not None:

                hyper_noise_map_1d = galaxy.hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                    noise_map=noise_map_1d,
                    hyper_model_image=galaxy.hyper_model_image_1d,
                    hyper_galaxy_image=galaxy.hyper_galaxy_image_1d,
                )

                hyper_noise_maps_1d.append(hyper_noise_map_1d)

            else:

                hyper_noise_maps_1d.append(None)

        return hyper_noise_maps_1d


class Plane(AbstractDataPlane):
    def __init__(
        self,
        galaxies,
        grid_stack,
        redshift=None,
        border=None,
        compute_deflections=True,
        cosmology=cosmo.Planck15,
    ):
        """A plane of galaxies where all galaxies are at the same redshift.

        Parameters
        -----------
        redshift : float or None
            The redshift of the plane.
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        grid_stack : masks.GridStack
            The stack of grid_stacks of (y,x) arc-second coordinates of this plane.
        border : masks.RegularGridBorder
            The borders of the regular-grid, which is used to relocate demagnified traced regular-pixel to the \
            source-plane borders.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        cosmology : astropy.cosmology
            The cosmology associated with the plane, used to convert arc-second coordinates to physical values.
        """

        if redshift is None:

            if not galaxies:
                raise exc.RayTracingException(
                    "A redshift and no galaxies were input to a Plane. A redshift for the Plane therefore cannot be"
                    "determined"
                )
            elif not all(
                [galaxies[0].redshift == galaxy.redshift for galaxy in galaxies]
            ):
                raise exc.RayTracingException(
                    "A redshift and two or more galaxies with different redshifts were input to a Plane. A unique "
                    "Redshift for the Plane therefore cannot be determined"
                )
            else:
                redshift = galaxies[0].redshift

        super(Plane, self).__init__(
            redshift=redshift,
            galaxies=galaxies,
            grid_stack=grid_stack,
            border=border,
            compute_deflections=compute_deflections,
            cosmology=cosmology,
        )


class PlanePositions(object):
    def __init__(
        self, redshift, galaxies, positions, compute_deflections=True, cosmology=None
    ):
        """A plane represents a set of galaxies at a given redshift in a ray-tracer_normal and the positions of image-plane \
        coordinates which mappers close to one another in the source-plane.

        Parameters
        -----------
        galaxies : [Galaxy]
            The list of lens galaxies in this plane.
        positions : [[[]]]
            The (y,x) arc-second coordinates of image-plane pixels which (are expected to) mappers to the same
            location(s) in the final source-plane.
        compute_deflections : bool
            If true, the deflection-angles of this plane's coordinates are calculated use its galaxy's mass-profiles.
        """

        self.redshift = redshift
        self.galaxies = galaxies
        self.positions = positions

        if compute_deflections:

            def calculate_deflections(pos):
                return sum(
                    map(lambda galaxy: galaxy.deflections_from_grid(pos), galaxies)
                )

            self.deflections = list(
                map(lambda pos: calculate_deflections(pos), self.positions)
            )

        self.cosmology = cosmology

    def trace_to_next_plane(self):
        """Trace the positions to the next plane."""
        return list(
            map(
                lambda positions, deflections: np.subtract(positions, deflections),
                self.positions,
                self.deflections,
            )
        )


class PlaneImage(scaled_array.ScaledRectangularPixelArray):
    def __init__(self, array, pixel_scales, grid, origin=(0.0, 0.0)):
        self.grid = grid
        super(PlaneImage, self).__init__(
            array=array, pixel_scales=pixel_scales, origin=origin
        )
